-- vim: ts=2 sw=2 ai et fenc=utf8
local MAJOR, MINOR = "LibGroupInSpecT-1.0", tonumber (("$Rev: 0$"):match ("(%d+)"))

local lib = LibStub:NewLibrary (MAJOR, MINOR)
if not lib then return end

if lib.state then return end -- Already loaded

--[[
-- These events can be registered for using the regular CallbackHandler ways.
--
-- "GroupInSpecT_Add", guid, unit, info
-- "GroupInSpecT_Update", guid, unit, info
-- "GroupInSpecT_Remove, guid, nil
--
-- Where <info> is a table containing some or all of the following:
--   .guid
--   .name
--   .realm
--   .class
--   .class_localized
--   .class_id
--   .global_spec_id
--   .talents = {
--     [<tier>] = {
--       [column] = { .name, .texture, .selected, .available },
--       ...
--     },
--     ...
--   }
--
-- Functions for external use:
--
--   lib:QueuedInspections ()
--     Returns an array of GUIDs of outstanding inspects.
--
--   lib:StaleInspections ()
--     Returns an array of GUIDs for which the data has become stale and is
--     awaiting an update (no action required, the refresh happens internally).
--     Due to Blizzard exposing no events on (re/un)talent, there will be
--     frequent marking of inspect data as being stale.
--
--   lib:GetCachedInfo (guid)
--     Returns the cached info for the given GUID, if available, nil otherwise.
--     Information is cached for current group members only.
--
--   lib:GroupUnits ()
--     Returns an array with the set of unit ids for the current group.
--]]

local ADD_EVENT    = "GroupInSpecT_Add"
local UPDATE_EVENT = "GroupInSpecT_Update"
local REMOVE_EVENT = "GroupInSpecT_Remove"

local INSPECT_DELAY = 1.5
local INSPECT_STALE_DELAY = 5
local INSPECT_TIMEOUT = 10 -- If we get no notification within 10s, give up on unit
local MAX_VALID_SPEC_ID = 500



local function debug (...)
--@debug@
  --print (...)
--@end-debug@
end


lib.events = LibStub ("CallbackHandler-1.0"):New (lib)

local function OnSpecUpdate (event, guid, info)
  local group = lib.state.group
  if group and group[guid] then
    debug ("Saved info for group member with guid "..guid)
    group[guid] = info
  end
end
lib:RegisterCallback (UPDATE_EVENT, OnSpecUpdate)


-- Frame for events
local frame = CreateFrame ("Frame", MAJOR .. "_Frame")
lib.frame = frame
frame:Hide()
frame:UnregisterAllEvents ()
frame:RegisterEvent ("PLAYER_ENTERING_WORLD")
frame:RegisterEvent ("PLAYER_LEAVING_WORLD")
frame:RegisterEvent ("INSPECT_READY")
frame:RegisterEvent ("GROUP_ROSTER_UPDATE")
frame:RegisterEvent ("UNIT_LEVEL")
frame:RegisterEvent ("PLAYER_TALENT_UPDATE")
frame:RegisterEvent ("PLAYER_SPECIALIZATION_CHANGED")
frame:RegisterEvent ("UNIT_SPELLCAST_SUCCEEDED")
frame:RegisterEvent ("UNIT_NAME_UPDATE")
frame:SetScript ("OnEvent", function (this, event, ...)
  return lib[event] (lib, ...)
end)


-- Hide our run-state in an easy-to-dump object
lib.state = {
  mainq = {}, staleq = {}, -- inspect queues
  t = 0,
  last_inspect = 0,
  current_guid = nil,
  group = {}
}


-- Inspects only work after entering the world, so track that
lib.state.logged_in = IsLoggedIn ()
function lib:PLAYER_ENTERING_WORLD ()
  lib.state.logged_in = true
  local guid = UnitGUID ("player")
  local info = self:BuildInfo ("player")
  self.events:Fire (ADD_EVENT, guid, "player", info)
  self.events:Fire (UPDATE_EVENT, guid, "player", info)
end

function lib:PLAYER_LEAVING_WORLD ()
  lib.state.logged_in = false
end


-- Simple timer
do
  lib.state.t = 0
  frame:SetScript("OnUpdate", function(this, elapsed)
    lib.state.t = lib.state.t + elapsed
    local timeout = next (lib.state.mainq) and INSPECT_DELAY or INSPECT_STALE_DELAY
    if lib.state.t > timeout then
      lib:ProcessQueues ()
      lib.state.t = 0
    end
  end)
end


-- Internal library functions

local unitstrings = {
  raid = { "player" }, -- This seems to be needed under certain circumstances. Odd.
  party = { "player" }, -- Player not part of partyN
  player = { "player" }
}
for i = 1,40 do table.insert (unitstrings.raid, "raid"..i) end
for i = 1,4  do table.insert (unitstrings.party, "party"..i) end

-- Returns an array with the set of unit ids for the current group
function lib:GroupUnits ()
  local units
  if IsInRaid () then
    units = unitstrings.raid
  elseif GetNumSubgroupMembers () > 0 then
    units = unitstrings.party
  else
    units = unitstrings.player
  end
  return units
end

function lib:GuidToUnit (guid)
  for i,unit in ipairs (self:GroupUnits ()) do
    if UnitExists (unit) and UnitGUID (unit) == guid then
      return unit
    end
  end
end



function lib:Query (unit)
  if not UnitIsPlayer (unit) then return end -- NPC

  if UnitIsUnit (unit, "player") then
    self.events:Fire (UPDATE_EVENT, UnitGUID("player"), "player", self:BuildInfo ("player"))
    return
  end

  local mainq, staleq = self.state.mainq, self.state.staleq

  local guid = UnitGUID (unit)
  if not mainq[name] then
    mainq[guid] = 1
    staleq[guid] = nil
    self.frame:Show () -- Start timer if not already running
  end
end


function lib:Refresh (unit)
  local guid = UnitGUID (unit)
  if not guid then return end
  debug ("Refreshing "..unit)
  if not self.state.mainq[guid] then
    self.state.staleq[guid] = 1
    self.frame:Show ()
  end
end


function lib:ProcessQueues ()
  if not self.state.logged_in then return end
  if InCombatLockdown () then return end -- Never inspect while in combat

  -- Don't' mess with the UI's inspections
	if InspectFrame and InspectFrame:IsShown () then
		return
	end

  local mainq = self.state.mainq
  local staleq = self.state.staleq

  local q = mainq
  if not next (q) then q = staleq end

  if (self.state.last_inspect + INSPECT_TIMEOUT) < GetTime () then
    -- If there was an inspect going, it's timed out so kick it out of the queue
    local guid = self.state.current_guid
    if guid then
      debug ("Inspect timed out for "..guid)
      mainq[guid], staleq[guid] = nil, nil
      self.state.current_guid = nil
    end
  end

  if self.state.current_guid then return end -- Still waiting on our inspect data

	for guid,_ in pairs (q) do
		local unit = self:GuidToUnit (guid)
		if not unit or not CanInspect (unit) then
			q[guid] = nil
		else
      debug ("Inspecting "..unit..", aka "..UnitName(unit))
      self.state.current_guid = guid
		  NotifyInspect (unit)
      break
    end
	end

  if not next (self.state.mainq) and not next (self.state.staleq) then
    frame:Hide() -- Cancel timer, nothing queued
  end
end


hooksecurefunc("NotifyInspect", function (...) return lib:NotifyInspect (...) end)
function lib:NotifyInspect(unit)
	self.state.last_inspect = GetTime()
end

function lib:BuildInfo (unit)
  local info = {}
  info.guid = guid
  info.name, info.realm = UnitName (unit)
  if info.realm and info.realm == "" then info.realm = nil end
  info.class_localized, info.class, info.class_id = UnitClass (unit)

  local is_inspect = not UnitIsUnit("player", unit)
  info.global_spec_id = is_inspect and GetInspectSpecialization (unit) or GetSpecializationInfo (GetSpecialization (nil) )
  if not info.global_spec_id or info.global_spec_id > MAX_VALID_SPEC_ID then
    -- Can't get valid spec/talent info at this point
    info.global_spec_id = nil
    return info
  end

  info.talents = {}
  local idx = 1
  local tname, tex, tier, column, sel, avail = GetTalentInfo (idx, is_inspect, nil, unit, info.class_id)
  while tier and column do
    if not info.talents[tier] then info.talents[tier] = {} end
    info.talents[tier][column] = {
      name = tname,
      texture = tex,
      selected = sel,
      available = avail
    }
    idx = idx + 1
    tname, tex, tier, column, sel, avail = GetTalentInfo (idx, is_inspect, nil, unit, info.class_id)
  end
  return info
end

function lib:INSPECT_READY (guid)
	local unit = self:GuidToUnit (guid)
	if unit then
    if guid == self.state.current_guid then
      self.state.current_guid = nil -- Got what we asked for
      debug("Got inspection data for requested guid "..guid)
    end

    local mainq, staleq = self.state.mainq, self.state.staleq
    mainq[guid], staleq[guid] = nil, nil

    local gspec_id = GetInspectSpecialization (unit)
    if gspec_id > MAX_VALID_SPEC_ID then -- Bah, got garbage, flag as stale and try again
      self.state.staleq[guid] = guid
      return
    end

    self.events:Fire (UPDATE_EVENT, guid, unit, self:BuildInfo (unit))
	end
end


-- Group handling parts

function lib:GROUP_ROSTER_UPDATE ()
  local group = self.state.group
  local members = {}
  local units = self:GroupUnits ()
  -- Find new members
  for i,unit in ipairs (self:GroupUnits ()) do
    local guid = UnitGUID (unit)
    if guid then
      members[guid] = 1
      if not group[guid] then
        local name = UnitName (unit)
        debug (name.." joined the group")
        group[guid] = 0
        self:Query (unit)
        self.events:Fire (ADD_EVENT, guid, unit, self:BuildInfo (unit))
      end
    end
  end
  -- Find removed members
  for guid,info in pairs (group) do
    if not members[guid] then
      debug (guid.." left the group")
      group[guid] = nil
      self.events:Fire (REMOVE_EVENT, guid, nil)
    end
  end
end

function lib:UNIT_LEVEL (unit)
  self:Refresh (unit)
end

function lib:PLAYER_TALENT_UPDATE ()
  self:Query ("player")
end

function lib:PLAYER_SPECIALIZATION_CHANGED ()
  self:Query ("player")
end

function lib:UNIT_NAME_UPDATE (unit)
  self.events:Fire (UPDATE_EVENT, UnitGUID (unit), unit, self:BuildInfo (unit))
end

local dual_spec_activations = {}
for i, spellid in ipairs (TALENT_ACTIVATION_SPELLS) do dual_spec_activations[spellid] = true end
function lib:UNIT_SPELLCAST_SUCCEEDED (unit, spellid)
  if dual_spec_activations[spellid] then
    self:Query (unit) -- Definitely changed, so high prio refresh
  end
end

-- Blizzard doesn't expose any events that indicate a talent/untalent/retalent event.
-- Until they do, we're stuck with intermittently rescanning everyone :(
-- I'm not convinced UNIT_AURA is the best hook to get this done, but it does have
-- the advantage of triggering on units coming into range, so we should have a
-- relatively fresh view of anyone around us.
function lib:UNIT_AURA (unit)
  local units = self:GroupUnits ()
  for i,gu in ipairs (self:GroupUnits ()) do
    if unit == gu then
      self:Refresh (unit)
   end
 end
end



-- External library functions

function lib:QueuedInspections ()
  local q = {}
  for guid,_ in pairs (self.state.mainq) do
    table.insert (q, guid)
  end
  return q
end

function lib:StaleInspections ()
  local q = {}
  for guid,_ in pairs (self.state.staleq) do
    table.insert (q, guid)
  end
  return q
end

function lib:GetCachedInfo (guid)
  local group = self.state.group
  if guid and group[guid] and group[guid] ~= 0 then
    return group[guid]
  else
    return nil
  end
end

